@@ -538,8 +538,6 @@ class Simulation(Frame):
             self.gas.Sigma = Field(self, SigmaGas,
                                    description="Surface density [g/cm²]")
             self.gas.Sigma.jacobinator = std.gas.jacobian
-        # Hidden fields required for the calculation of omplicit quantities
-        self.gas._rhs = np.zeros(shape1)
         # Temperature
         if self.gas.T is None:
             self.gas.T = Field(self, np.zeros(shape1),
@@ -932,6 +930,7 @@ class Simulation(Frame):
                                    description="Dust: explicit 5th-order adaptive Cash-Karp method"
                                    )
                 self.integrator.instructions[i] = inst
+                self.t.suggest(1.*c.year)
 
             else:
                 raise RuntimeError("Invalid method for explicit integration.")
