@@ -1,6 +1,7 @@
 '''Module containing standard functions for the gas.'''
 
 import numpy as np
+import scipy.sparse as sp
 from simframe.integration import Scheme
 
 import dustpy.constants as c
@@ -67,9 +68,10 @@ def set_implicit_boundaries(sim):
     sim : Frame
         Parent simulation frame"""
     # Total source terms
-    sim.gas.S.tot[0] = (sim.gas.Sigma[0]-sim.gas._SigmaOld[0])/sim.t.stepsize
+    sim.gas.S.tot[0] = (sim.gas.Sigma[0]-sim.gas._SigmaOld[0]
+                        )/(sim.t.prevstepsize+1.e-100)
     sim.gas.S.tot[-1] = (sim.gas.Sigma[-1] -
-                         sim.gas._SigmaOld[-1])/sim.t.stepsize
+                         sim.gas._SigmaOld[-1])/(sim.t.prevstepsize+1.e-100)
     # Hydrodynamic source terms
     sim.gas.S.hyd[0] = sim.gas.S.tot[0]
     sim.gas.S.hyd[-1] = sim.gas.S.tot[-1]
@@ -80,6 +82,31 @@ def set_implicit_boundaries(sim):
                       * (sim.grid.ri[-1]**2-sim.grid.ri[-2]**2))/sim.grid.ri[-1]
 
 
+def dt(sim):
+    """Function calculates the time step from the gas sources.
+
+    Parameters
+    ----------
+    sim : Frame
+        Parent simulation frame
+
+    Returns
+    -------
+    dt : float
+        Gas time step"""
+    if np.any(sim.gas.S.tot[1:-1] < 0.):
+        mask = np.logical_and(
+            sim.gas.Sigma > sim.gas.SigmaFloor,
+            sim.gas.S.tot < 0.)
+        mask[0] = False
+        mask[-1:] = False
+        rate = sim.gas.Sigma[mask] / sim.gas.S.tot[mask]
+        try:
+            return np.min(np.abs(rate))
+        except:
+            return None
+
+
 def cs_adiabatic(sim):
     """Function calculates the adiabatic sound speed of the gas. For the isothermal sound speed set
     ``Simulation.gas.gamma = 1``.
@@ -143,6 +170,28 @@ def Hp(sim):
 
 
 def jacobian(sim, x, *args, **kwargs):
+    """Functions calculates the Jacobian for the gas.
+
+    Parameters
+    ----------
+    sim : Frame
+        Parent simulation frame
+    x : IntVar
+        Integration variable
+    args : additional positional arguments
+    kwargs : additional keyworda arguments
+
+    Returns
+    -------
+    jac : Field
+        Jacobi matrix for gas evolution
+
+    Notes
+    -----
+    The boundaries need information about the time step, which is only available at
+    the integration stage. The boundary values are therefore meaningless and should not
+    be used to calculate the source terms via matrix-vector multiplication.
+    See the documentation for details."""
 
     # Parameters
     nu = sim.gas.nu * sim.dust.backreaction.A
@@ -152,32 +201,37 @@ def jacobian(sim, x, *args, **kwargs):
     dt = sim.t.stepsize
     r = sim.grid.r
     ri = sim.grid.ri
+    area = sim.grid.A
+    Nr = int(sim.grid.Nr)
 
     # Construct Jacobian
-    A, B, C = gas_f.jac_abc(nu, r, ri, v)
-    jac = np.diag(A[1:], -1) + np.diag(B) + np.diag(C[:-1], 1)
-
-    # Boundaries
+    A, B, C = gas_f.jac_abc(area, nu, r, ri, v)
+    J = sp.diags(
+        (A[1:], B[:], C[:-1]),
+        offsets=(-1, 0, 1),
+        shape=(Nr, Nr),
+        format="csc")
 
     # Right hand side
     sim.gas._rhs[:] = sim.gas.Sigma
 
+    # Boundaries
+    B00, B01, B02 = 0., 0., 0.
+    BNm1Nm3, BNm1Nm2, BNm1Nm1 = 0., 0., 0.
+
     # Inner boundary
     if sim.gas.boundary.inner is not None:
         # Given value
         if sim.gas.boundary.inner.condition == "val":
-            jac[0, 0] = 0.
             sim.gas._rhs[0] = sim.gas.boundary.inner.value
         # Constant value
         elif sim.gas.boundary.inner.condition == "const_val":
-            jac[0, 0] = 0.
-            jac[0, 1] = 1./dt
+            B01 = (1./dt)[0]
             sim.gas._rhs[0] = 0.
         # Given gradient
         elif sim.gas.boundary.inner.condition == "grad":
             K1 = - r[1]/r[0]
-            jac[0, 0] = 0.
-            jac[0, 1] = -K1/dt
+            B01 = -(K1/dt)[0]
             sim.gas._rhs[0] = - ri[1]/r[0] * \
                 (r[1]-r[0])*sim.gas.boundary.inner.value
         # Constant gradient
@@ -185,41 +239,39 @@ def jacobian(sim, x, *args, **kwargs):
             Di = ri[1]/ri[2] * (r[1]-r[0]) / (r[2]-r[0])
             K1 = - r[1]/r[0] * (1. + Di)
             K2 = r[2]/r[0] * Di
-            jac[0, 0] = 0.
-            jac[0, 1] = -K1/dt
-            jac[0, 2] = -K2/dt
+            B00, B01, B02 = 0., -(K1/dt)[0], -(K2/dt)[0]
             sim.gas._rhs[0] = 0.
         # Given power law
         elif sim.gas.boundary.inner.condition == "pow":
             p = sim.gas.boundary.inner.value
-            jac[0, 0] = 0.
             sim.gas._rhs[0] = sim.gas.Sigma[1] * (r[0]/r[1])**p
         # Constant power law
         elif sim.gas.boundary.inner.condition == "const_pow":
             p = np.log(sim.gas.Sigma[2] /
                        sim.gas.Sigma[1]) / np.log(r[2]/r[1])
             K1 = - (r[0]/r[1])**p
-            jac[0, 0] = 0.
-            jac[0, 1] = -K1/dt
+            B01 = -(K1/dt)[0]
             sim.gas._rhs[0] = 0.
 
+    row = [0, 0, 0]
+    col = [0, 1, 2]
+    dat = [B00, B01, B02]
+    gen = (dat, (row, col))
+    J_in = sp.csc_matrix(gen, shape=(Nr, Nr))
+
     # Outer boundary
-    # Right hand side
     if sim.gas.boundary.outer is not None:
         # Given value
         if sim.gas.boundary.outer.condition == "val":
-            jac[-1, -1] = 0
             sim.gas._rhs[-1] = sim.gas.boundary.outer.value
         # Constant value
         elif sim.gas.boundary.outer.condition == "const_val":
-            jac[-1, -1] = 0
-            jac[-1, -2] = 1./dt
+            BNm1Nm2 = (1./dt)[0]
             sim.gas._rhs[-1] = 0.
         # Given gradient
         elif sim.gas.boundary.outer.condition == "grad":
             KNrm2 = - r[-2]/r[-1]
-            jac[-1, -1] = 0.
-            jac[-1, -2] = -KNrm2/dt
+            BNm1Nm2 = -(KNrm2/dt)[0]
             sim.gas._rhs[-1] = ri[-2]/r[-1] * \
                 (r[-1]-r[-2])*sim.gas.boundary.outer.value
         # Constant gradient
@@ -227,25 +279,28 @@ def jacobian(sim, x, *args, **kwargs):
             Do = ri[-2]/ri[-3] * (r[-1]-r[-2]) / (r[-2]-r[-3])
             KNrm2 = - r[-2]/r[-1] * (1. + Do)
             KNrm3 = r[-3]/r[-1] * Do
-            jac[-1, -1] = 0.
-            jac[-1, -2] = -KNrm2/dt
-            jac[-1, -3] = -KNrm3/dt
+            Bnm1Bnm2 = -(KNrm2/dt)[0]
+            Bnm1Bnm3 = -(KNrm3/dt)[0]
             sim.gas._rhs[-1] = 0.
         # Given power law
         elif sim.gas.boundary.outer.condition == "pow":
             p = sim.gas.boundary.outer.value
-            jac[-1, -1] = 0.
             sim.gas._rhs[-1] = sim.gas.Sigma[-2] * (r[-1]/r[-2])**p
         # Constant power law
         elif sim.gas.boundary.outer.condition == "const_pow":
             p = np.log(sim.gas.Sigma[-2] /
-                       sim.gas.Sigma[-3]) / np.log(r[-2]/r[-1])
+                       sim.gas.Sigma[-3]) / np.log(r[-2]/r[-3])
             KNrm2 = - (r[-1]/r[-2])**p
-            jac[-1, -1] = 0.
-            jac[-1, -2] = -KNrm2/dt
+            Bnm1Bnm2 = -(KNrm2/dt)[0]
             sim.gas._rhs[-1] = 0.
 
-    return jac
+    row = [Nr-1, Nr-1, Nr-1]
+    col = [Nr-3, Nr-2, Nr-1]
+    dat = [BNm1Nm3, BNm1Nm2, BNm1Nm1]
+    gen = (dat, (row, col))
+    J_out = sp.csc_matrix(gen, shape=(Nr, Nr))
+
+    return J_in + J + J_out
 
 
 def lyndenbellpringle1974(r, rc, p, Mdisk):
@@ -259,7 +314,7 @@ def lyndenbellpringle1974(r, rc, p, Mdisk):
         critical cutoff radius
     p : float
         power law exponent
-    Mdist : float
+    Mdisk : float
         disk mass
 
     Returns
@@ -421,7 +476,7 @@ def vvisc(sim):
     return gas_f.v_visc(sim.gas.Sigma, sim.gas.nu, sim.grid.r, sim.grid.ri)
 
 
-def _f_impl_1_euler_direct(x0, Y0, dx, jac=None, rhs=None, *args, **kwargs):
+def _f_impl_1_direct(x0, Y0, dx, jac=None, rhs=None, *args, **kwargs):
     """Implicit 1st-order Euler integration scheme with direct matrix inversion
 
     Parameters
@@ -449,20 +504,24 @@ def _f_impl_1_euler_direct(x0, Y0, dx, jac=None, rhs=None, *args, **kwargs):
        | 1 
     """
     if jac is None:
-        jac = Y0.jacobian(x0 + dx, dt=dx)
+        jac = Y0.jacobian(x0 + dx)
     if rhs is None:
         rhs = np.array(Y0)
 
     # Add external source terms to right-hand side
-    rhs[1:-1] += dx*Y0._owner.gas.S.ext[1:-1]
+    rhs[1:-1] += dx[0]*Y0._owner.gas.S.ext[1:-1]
 
     N = jac.shape[0]
-    eye = np.eye(N)
-    A = eye - dx * jac
-    Y1 = np.linalg.inv(A) @ rhs
+    eye = sp.identity(N, format="csc")
+    A = eye - dx[0]*jac
+    A_LU = sp.linalg.splu(A,
+                          permc_spec="MMD_AT_PLUS_A",
+                          diag_pivot_thresh=0.0,
+                          options=dict(SymmetricMode=True))
+    Y1 = A_LU.solve(rhs)
 
     return Y1 - Y0
 
 
-impl_1_euler_direct = Scheme(
-    _f_impl_1_euler_direct, description="Implicit 1st-order direct Euler method")
+impl_1_direct = Scheme(
+    _f_impl_1_direct, description="Implicit 1st-order direct solver")
